@@ -43,7 +43,8 @@ class Header:
     mask_filenames : str
         A list of masks for the slides.  Each slide's mask will be used to select which tiles of the
         slide to process.  If the mask does not have one pixel per tile then it will be upsampled or
-        downsampled as necessary.
+        downsampled as necessary.  An empty string indicates that no mask file is being supplied and
+        that all tiles should be retained.
 
     Notes
     -----
@@ -77,7 +78,9 @@ class Header:
                     self.dictionary[key] = self.dictionary[key] * len(filenames)
 
     def keys(self):
-        """The method that returns the keys of the key-value pairs stored by histomics_stream.dsm.wsi.Header."""
+        """The method that returns the keys of the key-value pairs stored by
+        histomics_stream.dsm.wsi.Header.
+        """
         return self.dictionary.keys()
 
     def __getitem__(self, key):
@@ -125,14 +128,19 @@ class Print:
 
     def __init__(self, member):
         self.member = member
-        tf.print("Running histomics_stream.dsm.wsi.Print.__init__, with member = ", self.member)
+        tf.print(
+            "Running histomics_stream.dsm.wsi.Print.__init__, with member = ",
+            self.member,
+        )
         print("Tracing histomics_stream.dsm.wsi.Print.__init__")
 
     @tf.function
     def __call__(self, elem):
         """This method is called by tensorflow to do the work of this class."""
-
-        tf.print("Running histomics_stream.dsm.wsi.Print.__call__, with member = ", self.member)
+        tf.print(
+            "Running histomics_stream.dsm.wsi.Print.__call__, with member = ",
+            self.member,
+        )
         print("Tracing histomics_stream.dsm.wsi.Print.__call__")
         return elem
 
@@ -169,15 +177,22 @@ class ComputeReadParameters:
             inp=[elem["filename"], elem["magnification"], self.tolerance],
             Tout=(tf.int32, tf.float32, tf.int32, tf.int32),
         )
-        response = {**elem, "level": level, "factor": factor, "width": width, "height": height}
+        response = {
+            **elem,
+            "level": level,
+            "factor": factor,
+            "width": width,
+            "height": height,
+        }
         return response
 
-    def _py_compute_read_parameters(self, filenameIn, magnificationIn, toleranceIn):
-        """This method is the internal py_function (i.e. not @tf.function) that does the actual work of this class."""
-
-        filename = filenameIn.numpy().decode("utf-8")
-        magnification = magnificationIn.numpy()
-        tolerance = toleranceIn.numpy()
+    def _py_compute_read_parameters(self, filename_in, magnification_in, tolerance_in):
+        """This method is the internal py_function (i.e. not @tf.function) that does the actual work of
+        this class.
+        """
+        filename = filename_in.numpy().decode("utf-8")
+        magnification = magnification_in.numpy()
+        tolerance = tolerance_in.numpy()
 
         if re.compile(r"\.svs$").search(filename):
             # read whole-slide image file and create openslide object
@@ -247,16 +262,15 @@ class ComputeReadParameters:
 
 
 class AddTileDescription:
-    """A class for supplying tile size and other information that can be supplied to tensorflow.dataset.map
+    """A class for supplying tile size and other information that can be supplied to
+    tensorflow.dataset.map.
 
     An instance of class histomics_stream.dsm.wsi.AddTileDescription can be supplied as an argument to
     tensorflow.dataset.map.  histomics_stream.dsm.wsi.AddTileDescription adds new key-value pairs to the
     tensorflow dictionary to set the desired tile width, height, width overlap, and height overlap for
-    each element, and to indicate whether we want tiles even if they are fractional (aka of truncated
-    size) due to partially falling off the edge of the image.  (These fractional tiles can be
-    problematic because tensorflow likes its shapes to be uniform.)  chunk_width_factor and
-    chunk_height_factor indicate how many tiles are read at a time.  The primary functionality of this
-    class over an ordinary dictionary is that it sets all the required keys and no others.
+    each element.  chunk_width_factor and chunk_height_factor indicate how many tiles are read at a
+    time.  The primary functionality of this class over an ordinary dictionary is that it sets all the
+    required keys and no others.
 
     Parameters
     ----------
@@ -272,8 +286,6 @@ class AddTileDescription:
         The width of a chunk read from disk at one time as measured in number of (possibly) tiles.
     chunk_height_factor : tf.constant(, dtype=tf.int32)
         The height of a chunk read from disk at one time as measured in number of (possibly) tiles.
-    fractional : tf.constant(, dtype=tf.bool)
-        Set to True if tiles at the edges of the slide that are less than full width or full height should be retained.
 
     Notes
     -----
@@ -291,7 +303,6 @@ class AddTileDescription:
         overlap_height=tf.constant(0, dtype=tf.int32),
         chunk_width_factor=tf.constant(8, dtype=tf.int32),
         chunk_height_factor=tf.constant(8, dtype=tf.int32),
-        fractional=tf.constant(False, dtype=tf.bool),
     ):
         self.dictionary = {
             "tw": tile_width,
@@ -300,7 +311,6 @@ class AddTileDescription:
             "oh": overlap_height,
             "cwf": chunk_width_factor,
             "chf": chunk_height_factor,
-            "fractional": fractional,
         }
 
     @tf.function
@@ -319,9 +329,7 @@ class ComputeResampledMask:
     tile in the input image.  Note that we are assuming that this will take care of any aspects related
     to the overlapping of tiles.  Subsequent to that, we will not be looking at the mask pixels for
     adjacent tiles even though they may overlap with the tile being considered.  Note further that we
-    are assuming that the mask will be downsampled (or upsampled) to have one whole pixel per tile, even
-    if not every tile is whole; that is even if some are fractional tiles from the right or bottom
-    edges.
+    are assuming that the mask will be downsampled (or upsampled) to have one whole pixel per tile.
 
     """
 
@@ -329,59 +337,77 @@ class ComputeResampledMask:
     def __call__(self, elem):
         """This method is called by tensorflow to do the work of this class."""
 
-        if "mask_filename" in elem.keys():
-            mask_wsi = tf.py_function(
-                func=self._py_compute_resampled_mask,
-                inp=[
-                    elem["mask_filename"],
-                    elem["width"],
-                    elem["height"],
-                    elem["cwf"],
-                    elem["chf"],
-                    elem["tw"],
-                    elem["th"],
-                    elem["ow"],
-                    elem["oh"],
-                    elem["fractional"],
-                ],
-                Tout=tf.bool,
-            )
-            return {**elem, "mask_wsi": mask_wsi}
-        else:
-            return elem
+        mask_wsi = tf.py_function(
+            func=self._py_compute_resampled_mask,
+            inp=[
+                elem["mask_filename"],
+                elem["width"],
+                elem["height"],
+                elem["cwf"],
+                elem["chf"],
+                elem["tw"],
+                elem["th"],
+                elem["ow"],
+                elem["oh"],
+            ],
+            Tout=tf.uint8,
+        )
+        return {**elem, "mask_wsi": mask_wsi}
 
     def _py_compute_resampled_mask(
-        self, mask_filenameIn, widthIn, heightIn, cwfIn, chfIn, twIn, thIn, owIn, ohIn, fractionalIn
+        self,
+        mask_filename_in,
+        width_in,
+        height_in,
+        cwf_in,
+        chf_in,
+        tw_in,
+        th_in,
+        ow_in,
+        oh_in,
     ):
-        """This method is the internal py_function (i.e. not @tf.function) that does much of the actual work of this class."""
+        """This method is the internal py_function (i.e. not @tf.function) that does much of the actual
+        work of this class.
+        """
 
-        mask_filename = mask_filenameIn.numpy().decode("utf-8")
-        width = widthIn.numpy()
-        height = heightIn.numpy()
-        cwf = cwfIn.numpy()
-        chf = chfIn.numpy()
-        tw = twIn.numpy()
-        th = thIn.numpy()
-        ow = owIn.numpy()
-        oh = ohIn.numpy()
-        fractional = fractionalIn.numpy()
+        mask_filename = mask_filename_in.numpy().decode("utf-8")
+        width = width_in.numpy()
+        height = height_in.numpy()
+        cwf = cwf_in.numpy()
+        chf = chf_in.numpy()
+        tw = tw_in.numpy()
+        th = th_in.numpy()
+        ow = ow_in.numpy()
+        oh = oh_in.numpy()
 
-        left_bound = max(0, width - ow if fractional else width - tw + 1)
-        top_bound = max(0, height - oh if fractional else height - th + 1)
+        left_bound = max(0, width - tw + 1)
+        top_bound = max(0, height - th + 1)
 
-        # Read in an image from the supplied file name for the mask
-        mask_itk = itk.imread(mask_filename)
-        if mask_itk.GetImageDimension() != 2:
-            raise ValueError("The mask should be a 2-dimensional image.")
-        mask = tf.constant(itk.array_from_image(mask_itk))
-        # Add batch and channels dimensions
-        mask = mask[tf.newaxis, ..., tf.newaxis]
-        # The new size is one pixel per tile
+        # The desired mask size is one pixel per tile
         resampled_width = int(np.floor((left_bound - 1) / (tw - ow)) + 1)
         resampled_height = int(np.floor((top_bound - 1) / (th - oh)) + 1)
-        if abs(math.log((resampled_width / mask.shape[2]) / (resampled_height / mask.shape[1]))) > 0.20:
-            raise ValueError("The mask aspect ratio does not match the image aspect ratio.")
         resampled_shape = (resampled_height, resampled_width)
+        # By default assume that all tiles will be retained.
+        mask = tf.convert_to_tensor(
+            np.ones((resampled_height, resampled_width, 1), dtype=np.uint8),
+            dtype=tf.uint8,
+        )
+
+        if mask_filename != "":
+            # Read in an image from the supplied file name for the mask
+            mask_itk = itk.imread(mask_filename)
+            if mask_itk.GetImageDimension() != 2:
+                raise ValueError("The mask should be a 2-dimensional image.")
+            mask = tf.constant(itk.array_from_image(mask_itk), dtype=tf.uint8)
+
+            # Add batch and channels dimensions
+            mask = mask[tf.newaxis, ..., tf.newaxis]
+            if (
+                abs(math.log((resampled_width / mask.shape[2]) / (resampled_height / mask.shape[1])))
+                > 0.20
+            ):
+                raise ValueError("The mask aspect ratio does not match the image aspect ratio.")
+
         # Perform the resampling
         resampled = tf.image.resize(mask, resampled_shape)[0, ...]
 
@@ -390,16 +416,16 @@ class ComputeResampledMask:
         # multiples of those values.
         # * Also, we reshape so that the mask has shape (padded_height, padded_width, channels) where
         # channels = 1.
-        # * Also, we cast the array to type np.bool.  (Note that we use a formula with floor() rather than a
-        # formula with ceil() so that we get the same answer with float or integer division for the
-        # argument.)
+        # * Also, we cast the array to type np.uint8.  (Note that we use a formula with floor() rather
+        # than a formula with ceil() so that we get the same answer with float or integer division for
+        # the argument.)
         padded_resampled = np.zeros(
             (
                 int((tf.math.floor((resampled_shape[0] - 1) / chf) + 1) * chf),
                 int((tf.math.floor((resampled_shape[1] - 1) / cwf) + 1) * cwf),
                 1,
             ),
-            dtype=np.bool,
+            dtype=np.uint8,
         )
         padded_resampled[
             : resampled_shape[0],
@@ -411,7 +437,8 @@ class ComputeResampledMask:
         # print(f"resampled_shape = {resampled_shape}")
         # print(f"padded_resampled.shape = {padded_resampled.shape}")
         # tf.print(padded_resampled[:8,:8,0])
-        return padded_resampled
+        response = tf.convert_to_tensor(padded_resampled, dtype=tf.uint8)
+        return response
 
 
 class ComputeChunkPositions:
@@ -419,9 +446,9 @@ class ComputeChunkPositions:
 
     An instance of class histomics_stream.dsm.wsi.ComputeChunkPositions can be supplied as an argument
     to tensorflow.dataset.map.  histomics_stream.dsm.wsi.ComputeChunkPositions figures out what the read
-    chunks will be based upon the tile parameters (size, overlap, fractional).  It divvys up the mask
-    into pieces corresponding to the read chunks.  Note that it is important to subsequently call
-    .unbatch() when it is desired that the chunks be not batched by slide.
+    chunks will be based upon the tile parameters (size, overlap).  It divvys up the mask into pieces
+    corresponding to the read chunks.  Note that it is important to subsequently call .unbatch() when it
+    is desired that the chunks be not batched by slide.
 
     """
 
@@ -434,19 +461,13 @@ class ComputeChunkPositions:
         chunk_width = elem["cwf"] * (elem["tw"] - elem["ow"]) + elem["ow"]
         chunk_height = elem["chf"] * (elem["th"] - elem["oh"]) + elem["oh"]
 
-        # The left side of a tile cannot be as large as left_bound.  Also, the left side of a chunk cannot
-        # be as large as left_bound because chunks contain a whole number of tiles.
-        left_bound = tf.maximum(
-            zero,
-            elem["width"] - elem["ow"] if elem["fractional"] else elem["width"] - elem["tw"] + one,
-        )
+        # The left side of a tile cannot be as large as left_bound.  Also, the left side of a chunk
+        # cannot be as large as left_bound because chunks contain a whole number of tiles.
+        left_bound = tf.maximum(zero, elem["width"] - elem["tw"] + one)
         chunk_left = tf.range(zero, left_bound, chunk_width - elem["ow"])
         chunk_right = tf.clip_by_value(chunk_left + chunk_width, zero, elem["width"])
 
-        top_bound = tf.maximum(
-            zero,
-            elem["height"] - elem["oh"] if elem["fractional"] else elem["height"] - elem["th"] + one,
-        )
+        top_bound = tf.maximum(zero, elem["height"] - elem["th"] + one)
         chunk_top = tf.range(zero, top_bound, chunk_height - elem["oh"])
         chunk_bottom = tf.clip_by_value(chunk_top + chunk_height, zero, elem["height"])
 
@@ -454,51 +475,51 @@ class ComputeChunkPositions:
         cw = tf.tile(chunk_right - chunk_left, tf.stack([tf.size(chunk_top)]))
         cy = tf.repeat(chunk_top, tf.size(chunk_left))
         ch = tf.repeat(chunk_bottom - chunk_top, tf.size(chunk_left))
-        len = tf.size(cx)
+        chunk_len = tf.size(cx)
 
-        # If a mask was supplied, compute a mask for each chunk.  The size of a mask for a chunk will be
-        # chunk_width_factor by chunk_height_factor, even along the right or bottom border where it will be
-        # padded if necessary.
-        has_mask = "mask_wsi" in elem.keys()
+        # Compute a mask for each chunk.  The size of a mask for a chunk will be chunk_width_factor by
+        # chunk_height_factor, even along the right or bottom border where it will be padded if
+        # necessary.
 
-        mask_chunks = tf.TensorArray(dtype=tf.bool, size=len)
-        if has_mask:
-            mask_width = tf.shape(elem["mask_wsi"])[1]
-            mask_left = tf.cast(chunk_left / (elem["tw"] - elem["ow"]), dtype=tf.int32)
-            mask_right = tf.clip_by_value(mask_left + elem["cwf"], zero, mask_width)
-            mask_height = tf.shape(elem["mask_wsi"])[0]
-            mask_top = tf.cast(chunk_top / (elem["th"] - elem["oh"]), dtype=tf.int32)
-            mask_bottom = tf.clip_by_value(mask_top + elem["chf"], zero, mask_height)
-            mask_x = tf.tile(mask_left, tf.stack([tf.size(mask_top)]))
-            mask_w = tf.tile(mask_right - mask_left, tf.stack([tf.size(mask_top)]))
-            mask_y = tf.repeat(mask_top, tf.size(mask_left))
-            mask_h = tf.repeat(mask_bottom - mask_top, tf.size(mask_left))
-            len = tf.size(mask_x)
+        mask_chunks = tf.TensorArray(dtype=tf.uint8, size=chunk_len)
+        mask_width = tf.shape(elem["mask_wsi"])[1]
+        mask_left = tf.cast(chunk_left / (elem["tw"] - elem["ow"]), dtype=tf.int32)
+        mask_right = tf.clip_by_value(mask_left + elem["cwf"], zero, mask_width)
+        mask_height = tf.shape(elem["mask_wsi"])[0]
+        mask_top = tf.cast(chunk_top / (elem["th"] - elem["oh"]), dtype=tf.int32)
+        mask_bottom = tf.clip_by_value(mask_top + elem["chf"], zero, mask_height)
+        mask_x = tf.tile(mask_left, tf.stack([tf.size(mask_top)]))
+        mask_w = tf.tile(mask_right - mask_left, tf.stack([tf.size(mask_top)]))
+        mask_y = tf.repeat(mask_top, tf.size(mask_left))
+        mask_h = tf.repeat(mask_bottom - mask_top, tf.size(mask_left))
+        mask_len = tf.size(mask_x)
 
-            # mask_chunks = [elem["mask_wsi"][y:(y+h), x:(x+w)] for x, y, w, h in zip(mask_x, mask_w, mask_y, mask_h)]
-            condition = lambda i, _: tf.less(i, len)
-            body = lambda i, mask_chunks: (
-                i + 1,
-                mask_chunks.write(
-                    i,
-                    tf.image.crop_to_bounding_box(
-                        elem["mask_wsi"],
-                        tf.gather(mask_y, i),
-                        tf.gather(mask_x, i),
-                        tf.gather(mask_h, i),
-                        tf.gather(mask_w, i),
-                    ),
+        # mask_chunks = [elem["mask_wsi"][y:(y+h), x:(x+w)]
+        #                for x, y, w, h in zip(mask_x, mask_w, mask_y, mask_h)]
+        for i in tf.range(mask_len):
+            mask_chunks = mask_chunks.write(
+                i,
+                tf.image.crop_to_bounding_box(
+                    elem["mask_wsi"],
+                    tf.gather(mask_y, i),
+                    tf.gather(mask_x, i),
+                    tf.gather(mask_h, i),
+                    tf.gather(mask_w, i),
                 ),
             )
-            _, mask_chunks = tf.while_loop(condition, body, [0, mask_chunks])
 
         mask_chunks = mask_chunks.stack()
         response = {}
         for key in elem.keys():
             if key != "mask_wsi":
                 # Exclude mask_wsi because it is large and we now have mask_chunk.
-                response[key] = tf.repeat(elem[key], len)
-        response = {**response, "cx": cx, "cy": cy, "cw": cw, "ch": ch}
-        if has_mask:
-            response = {**response, "mask_chunk": mask_chunks}
+                response[key] = tf.repeat(elem[key], mask_len)
+        response = {
+            **response,
+            "cx": cx,
+            "cy": cy,
+            "cw": cw,
+            "ch": ch,
+            "mask_chunk": mask_chunks,
+        }
         return response
